class Issue:

    def __init__(self, title, _type = "INFORMATIONAL", description = ""):

        self.title = title
        self.description = description
        self.type = _type


    def as_dict(self):

        return {'title': self.title, 'description':self.description, 'type': self.type}


class Report:

    def __init__(self, issues = []):
        self.issues = issues


    def as_text(self):
        text = ""

        for issue in self.issues:
            text += "=== " + issue.title + " ===\n"
            text += "STATUS: " + issue.type + "\n"
            text += issue.description + "\n"

        return text




from z3 import *
from mythril.analysis.ops import *
from mythril.analysis.report import Issue


'''
MODULE DESCRIPTION:

Check for SUICIDE instructions that are not constrained by caller.
'''

def execute(statespace):

    issues = []

    svm = statespace.svm

    for k in svm.nodes:
        node = svm.nodes[k]

        for instruction in node.instruction_list:

            if(instruction['opcode'] == "SUICIDE"):

                state = node.states[instruction['address']]
                to = state.stack.pop()

                constraint_on_caller = False

                for constraint in node.constraints:
                    if "caller" in str(constraint):
                        constraint_on_caller = True
                        break

                if not constraint_on_caller:
                    s = Solver()

                    if (s.check() == sat):
                        issue = Issue("Unchecked SUICIDE", "VULNERABILITY")
                        issue.description = "The function " + node.function_name + " calls the SUICIDE instruction. It appears as if the function does not verify the caller address.\n"

                        if ("caller" in str(to)):
                            issue.description += "The remaining Ether is sent to the caller's address.\n"
                        elif ("storage" in str(to)):
                            issue.description += "The remaining Ether is sent to a stored address\n"
                        else:
                            issue.description += "The remaining Ether is sent to: " + str(to) + "\n"

                        issue.description += "Solver output:\n"

                        m = s.model()

                        for d in m.decls():
                            issue.description += "%s = 0x%x\n" % (d.name(), m[d].as_long())

                        issues.append(issue)    

    return issues






from laser.ethereum import svm
import logging
from .ops import *


class StateSpace:

    '''
    Symbolic EVM wrapper
    '''
    
    def __init__(self, contracts, dynloader = None, simplified=True):

        # Convert ETHContract objects to LASER SVM "modules"

        modules = {}

        for contract in contracts:
            modules[contract.address] = contract.as_dict()

        self.svm = svm.SVM(modules, simplified=True, dynamic_loader=dynloader)

        self.svm.sym_exec(contracts[0].address)

        self.modules = modules
        self.nodes = self.svm.nodes
        self.edges = self.svm.edges

        # Analysis

        self.calls = []
        self.suicides = []
        self.sstores = {}
        self.sloads = {}

        for key in self.svm.nodes:

            for instruction in self.nodes[key].instruction_list:

                op = instruction['opcode']

                if op in ('CALL', 'CALLCODE', 'DELEGATECALL', 'STATICCALL'):
                    stack = copy.deepcopy(self.svm.nodes[key].states[instruction['address']].stack)

                    if op in ('CALL', 'CALLCODE'):
                        gas, to, value, meminstart, meminsz, memoutstart, memoutsz = \
                            get_variable(stack.pop()), get_variable(stack.pop()), get_variable(stack.pop()), get_variable(stack.pop()), get_variable(stack.pop()), get_variable(stack.pop()), get_variable(stack.pop())

                        self.calls.append(Call(self.nodes[key], instruction['address'], op, to, value))
                    else:
                        gas, to, meminstart, meminsz, memoutstart, memoutsz = \
                            get_variable(stack.pop()), get_variable(stack.pop()), get_variable(stack.pop()), get_variable(stack.pop()), get_variable(stack.pop()), get_variable(stack.pop())

                        self.calls.append(Call(self.nodes[key], instruction['address'], op, to))
 

                    



import re
from z3 import *
import logging

TT256 = 2 ** 256
TT256M1 = 2 ** 256 - 1
TT255 = 2 ** 255


def safe_decode(hex_encoded_string):

    if (hex_encoded_string.startswith("0x")):
        return bytes.fromhex(hex_encoded_string[2:])
    else:
        return bytes.fromhex(hex_encoded_string)


def to_signed(i):
    return i if i < TT255 else i - TT256


def get_instruction_index(instruction_list, address):

	index = 0

	for instr in instruction_list:
		if instr['address'] == address:
			return index

		index += 1

	return None


def get_trace_line(instr, state):

	stack = str(state.stack[::-1])

	stack = re.sub("(\d+)",	lambda m: hex(int(m.group(1))), stack)
	stack = re.sub("\n", "", stack)

	return str(instr['address']) + " " + instr['opcode'] + "\tSTACK: " + stack


def pop_bitvec(state):
    # pop one element from stack, converting boolean expression to bitvector

    item = state.stack.pop()

    if type(item) == BoolRef:
        return If(item, BitVecVal(1, 256), BitVecVal(0, 256))
    else:
        return item


def get_concrete_int(item):

    if (type(item) == int):
        return item

    if (type(item) == BitVecNumRef):
        return item.as_long()

    return simplify(item).as_long()


def concrete_int_from_bytes(_bytes, start_index):

    b = _bytes[start_index:start_index+32]

    val = int.from_bytes(b, byteorder='big')

    return val


def concrete_int_to_bytes(val):

    logging.info("concrete_int_to_bytes " + str(bytes) + ", " + str(start_index))

    return (simplify(val).as_long()).to_bytes(32, byteorder='big')


def storage_constraints_for_path(svm, path):

    storage_constraints = []

    for edge in path:
       if (edge.condition is not None):

            cond = str(edge.condition)

            if 'storage' in cond:

                m = re.search(r'storage_([0-9a-f]+)', cond)

                if (m):

                    storage_constraints.append(m.group(1))

    return storage_constraints


def solver_for_path(svm, path):

    solver = Solver() 

    for edge in path:
        if edge.condition is not None:
            solver.add(edge.condition)

    return solver


def solve_path(svm, path):

    s = Solver()

    for edge in path:
        if edge.condition is not None:
            s.add(edge.condition)

    if (s.check() == sat):
        return s.model()

    else:
        return unsat


def satisfy_recursively(svm, node_addr, models = [], visited = []):

    if (node_addr in visited):
        logging.info("Circular reference, aborting")
        return None

    visited.append(node_addr)

    logging.info("Trying to solve constraints for node " + str(node_addr))

    for path in svm.paths[node_addr]:

        can_solve = True

        # Get constraints on storage locations

        constraints = storage_constraints_for_path(svm, path)

        nc = len(constraints)

        if (nc):

            logging.info("Path constrained by storage slots: " + str(constraints))

            logging.info("Trying to resolve " + str(nc) + " storage writes")

            solved = 0

            try:

                for storage_offset in constraints:

                    for _node_addr in svm.sstor_node_lists[storage_offset]:

                        m = satisfy_recursively(svm, _node_addr, models, visited)

                        logging.info("Satisfy returned " + str(m))

                        if m:
                            solved += 1
                            break

                logging.info(str(solved) + " of " + str(nc) + " writes satisfied")

                if solved == nc:

                    logging.info("Found viable path to node " + str(node_addr) + ", trying to solve")

                else:

                    logging.info("Unable to find viable path to node " + str(node_addr) + "")

                    can_solve = False

            except KeyError:
                    logging.info("No writes available to storage location")
                    can_solve = False

        else:

                logging.info("No storage constraints on path.")


        if can_solve:

            model = solve_path(svm, path)

            if (model == unsat):
                logging.info("Unsatisfiable")
                return False
            else:
                models.append(model)
                logging.info("Model found")               
                return True
                
        else:
            return False


def satisfy(svm, node_addr):

    models = []

    satisfy_recursively(svm, node_addr, models)

    return models


def debug_operand(operand):
    logging.info(str(type(operand)) + ", " + str(simplify(operand)))



from mythril.analysis.report import Report
from .modules import delegatecall_forward, unchecked_suicide


def fire_lasers(statespace):

	issues = []

	issues += delegatecall_forward.execute(statespace)
	issues += unchecked_suicide.execute(statespace)

	report = Report(issues)

	print(report.as_text())



from z3 import *
import re
import logging
from enum import Enum
from mythril.analysis.ops import *


class InputSource(Enum):
    CALLDATA = 1
    STORAGE = 2


def execute(statespace):

    for call in statespace.calls:

        logging.info("CALL: " + str(call.node.module_name) + ":" + str(call.node.function_name) + ": " + str(call.type) + " " + str(call.to) + " value " + str(call.value))

        interesting = False

        if (call.value.type == VarType.CONCRETE and call.value.val > 0) or (call.value.type == VarType.SYMBOLIC):

            logging.info("CALL with non-zero value: " + str(call.node.module_name) + ":" + str(call.node.function_name) + ": " + str(call.type) + " " + str(call.to) + " value " + str(call.value))

            interesting = True

            if call.to.type == VarType.CONCRETE:
                logging.info("Concrete 'to' address: " + call.to.hex())
            else:
                m = re.search(r"calldata", str(call.to))

                if (m):
                    logging.info("To-address from calldata: " + str(call.to))

            if call.value.type == VarType.CONCRETE:
                logging.info("Sending concrete amount: " + call.value)
            else:
                m = re.search(r"calldata", str(call.to))

                if (m):
                    logging.info("To-address from calldata: " + str(call.to))


        if (interesting):

            s = Solver()
            s.set(timeout=10000)

            for constraint in call.node.constraints:
                s.add(constraint)

            logging.info("SOLUTION:")

            if (s.check() == sat):

                m = s.model()

                for d in m.decls():
                    logging.info("%s = 0x%x" % (d.name(), m[d].as_long()))

            else:
                logging.info("unsat")



from z3 import *
import re
from mythril.analysis.ops import *
from mythril.analysis.report import Issue


'''
MODULE DESCRIPTION:

Check for invocations of delegatecall(msg.data) in the fallback function.
'''

def execute(statespace):

    issues = []

    for call in statespace.calls:

        if (call.type == "DELEGATECALL") and (call.node.function_name == "main"):

            stack = call.state.stack

            meminstart = get_variable(stack[-3])

            if meminstart.type == VarType.CONCRETE:

                if (re.search(r'calldata.*_0', str(call.state.memory[meminstart.val]))):

                    issue = Issue("CALLDATA forwarded with delegatecall()", "INFORMATION")
                    issue.description = \
                        "The contract '" +  str(call.node.module_name)  + "' forwards its calldata via DELEGATECALL in its fallback function. " \
                        "This means that any function in the called contract can be executed. Make sure to scan with the -l option so dependencies are checked as well.\n"      
                    
                    if (call.to.type == VarType.CONCRETE):
                        issue.description += ("DELEGATECALL target: " + hex(call.to.val))
                    else:
                        issue.description += "DELEGATECALL target: " + str(call.to)

                    issues.append(issue)

    return issues
